# Shamelessly stolen from Member Audit
from django.db import models
from esi.models import Token
from eveuniverse.models import EveSolarSystem, EveType

from allianceauth.eveonline.models import EveCharacter
from allianceauth.services.hooks import get_extension_logger
from app_utils.logging import LoggerAddTag

from .. import __title__
from ..app_settings import MININGTAXES_TAX_ONLY_CORP_MOONS
from ..decorators import fetch_token_for_character
from ..providers import esi
from .character import CharacterAbstract

logger = LoggerAddTag(get_extension_logger(__name__), __title__)


class AdminCharacter(CharacterAbstract):
    eve_character = models.OneToOneField(
        EveCharacter,
        related_name="miningtaxes_admin_character",
        on_delete=models.CASCADE,
    )

    @classmethod
    def get_esi_scopes(cls) -> list:
        return [
            "esi-industry.read_corporation_mining.v1",
            "esi-wallet.read_corporation_wallets.v1",
            "esi-universe.read_structures.v1",
        ]

    def update_all(self):
        if MININGTAXES_TAX_ONLY_CORP_MOONS:
            self.update_mining_observers()
        self.update_corp_ledger()
        return

    @fetch_token_for_character(
        ("esi-industry.read_corporation_mining.v1", "esi-universe.read_structures.v1")
    )
    def update_mining_observers(self, token: Token):
        logger.info("%s: Fetching mining observers from ESI", self)
        # get_corporation_corporation_id_mining_observers
        entries = esi.client.Industry.get_corporation_corporation_id_mining_observers(
            corporation_id=self.eve_character.corporation_id,
            token=token.valid_access_token(),
        ).results()
        for entry in entries:
            structinfo = esi.client.Universe.get_universe_structures_structure_id(
                structure_id=entry["observer_id"],
                token=token.valid_access_token(),
            ).results()
            structname = ""
            sys = None
            if type(structinfo) == dict:
                structname = structinfo["name"]
                sys, _ = EveSolarSystem.objects.get_or_create_esi(
                    id=structinfo["solar_system_id"]
                )
            else:
                logger.error("Wrong struct info for: %d" % entry["observer_id"])
                logger.error(structinfo)
                continue

            (obs, _) = self.mining_obs.update_or_create(
                obs_id=entry["observer_id"],
                defaults={
                    "obs_type": entry["observer_type"],
                    "sys_name": sys.name,
                    "name": structname,
                },
            )
            ledger = esi.client.Industry.get_corporation_corporation_id_mining_observers_observer_id(
                corporation_id=self.eve_character.corporation_id,
                observer_id=entry["observer_id"],
                token=token.valid_access_token(),
            ).results()
            for line in ledger:
                eve_type, _ = EveType.objects.get_or_create_esi(id=line["type_id"])
                obs.mining_log.update_or_create(
                    miner_id=line["character_id"],
                    date=line["last_updated"],
                    eve_type=eve_type,
                    defaults={
                        "eve_solar_system": sys,
                        "quantity": line["quantity"],
                    },
                )

    @fetch_token_for_character("esi-wallet.read_corporation_wallets.v1")
    def update_corp_ledger(self, token: Token):
        """Update corp ledger from ESI for this character."""
        logger.info("%s: Fetching corp wallet ledger from ESI", self)
        entries = (
            esi.client.Wallet.get_corporations_corporation_id_wallets_division_journal(
                corporation_id=self.eve_character.corporation_id,
                division=1,
                token=token.valid_access_token(),
            ).results()
        )
        for entry in entries:
            if entry["ref_type"] != "player_donation":
                continue
            self.corp_ledger.update_or_create(
                taxed_id=entry["first_party_id"],
                date=entry["date"],
                defaults={
                    "amount": entry["amount"],
                    "reason": entry["reason"][0:32],
                },
            )


class AdminMiningObservers(models.Model):
    """Mining Observers available to a character."""

    character = models.ForeignKey(
        AdminCharacter, on_delete=models.CASCADE, related_name="mining_obs"
    )
    obs_id = models.BigIntegerField()
    obs_type = models.CharField(max_length=32)
    name = models.CharField(max_length=32)
    sys_name = models.CharField(max_length=32)

    class Meta:
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=["obs_id"],
                name="functional_pk_mt_adminMiningObs",
            )
        ]

    def __str__(self) -> str:
        return f"{self.character} miningObs {self.id}"


class AdminMiningObsLog(models.Model):
    """Mining Log for a given Observer."""

    observer = models.ForeignKey(
        AdminMiningObservers, on_delete=models.CASCADE, related_name="mining_log"
    )
    date = models.DateField(db_index=True)
    miner_id = models.BigIntegerField()
    eve_type = models.ForeignKey(EveType, on_delete=models.CASCADE, related_name="+")
    quantity = models.PositiveIntegerField()
    observer_type = models.CharField(max_length=32)
    eve_solar_system = models.ForeignKey(
        EveSolarSystem, on_delete=models.CASCADE, related_name="+"
    )

    class Meta:
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=["observer", "date", "miner_id", "eve_type"],
                name="functional_pk_mt_adminMiningLog",
            )
        ]

    def __str__(self) -> str:
        return f"{self.observer} miningObs {self.id}"


class AdminMiningCorpLedgerEntry(models.Model):
    """Corp ledger entry of a character."""

    character = models.ForeignKey(
        AdminCharacter, on_delete=models.CASCADE, related_name="corp_ledger"
    )
    date = models.DateTimeField(db_index=True)
    taxed_id = models.BigIntegerField()
    amount = models.FloatField(default=0.0)
    reason = models.CharField(max_length=32)

    class Meta:
        default_permissions = ()
        constraints = [
            models.UniqueConstraint(
                fields=["character", "date", "taxed_id"],
                name="functional_pk_mt_admincorpledger",
            )
        ]

    def __str__(self) -> str:
        return f"{self.character} wallet {self.id}"



from app_utils.django import clean_setting

# put your app settings here

MININGTAXES_TAX_ONLY_CORP_MOONS = clean_setting("MININGTAXES_TAX_ONLY_CORP_MOONS", True)
"""Only tax corporate moons using moon observers as opposed to all moons appearing
in the personal mining ledgers.
"""

MININGTAXES_UPDATE_LEDGER_STALE = clean_setting("MININGTAXES_UPDATE_LEDGER_STALE", 240)
"""Minutes after which a character's mining ledger is considered stale
"""

MININGTAXES_UPDATE_STALE_OFFSET = clean_setting("MINGINGTAXES_UPDATE_STALE_OFFSET", 5)
"""Actual value for considering staleness of a ring will be the above value
minus this offset. Required to avoid time synchronization issues.
"""

MININGTAXES_TASKS_OBJECT_CACHE_TIMEOUT = clean_setting(
    "MEMBERAUDIT_TASKS_OBJECT_CACHE_TIMEOUT", 600
)

MININGTAXES_TASKS_TIME_LIMIT = clean_setting("MININGTAXES_TASKS_TIME_LIMIT", 7200)
"""Global timeout for tasks in seconds to reduce task accumulation during outages."""

MININGTAXES_REFINED_RATE = clean_setting("MININGTAXES_REFINED_RATE", 0.9063)
"""Refining rate for ores."""


MININGTAXES_PRICE_SOURCE_ID = clean_setting("MININGTAXES_PRICE_SOURCE_ID", 60003760)

MININGTAXES_PRICE_SOURCE_NAME = clean_setting("MININGTAXES_PRICE_SOURCE_NAME", "Jita")

MININGTAXES_PRICE_METHOD = clean_setting("MININGTAXES_PRICE_METHOD", "Fuzzwork")

MININGTAXES_PRICE_JANICE_API_KEY = clean_setting("MININGTAXES_PRICE_JANICE_API_KEY", "")



# Generated by Django 4.0.7 on 2022-11-08 03:06

from django.db import migrations


class Migration(migrations.Migration):

    dependencies = [
        ("miningtaxes", "0002_all_models"),
    ]

    operations = [
        migrations.AlterModelOptions(
            name="general",
            options={
                "default_permissions": (),
                "managed": False,
                "permissions": (
                    ("basic_access", "Can access this app"),
                    (
                        "admin_access",
                        "Can set tax rates for groups and add accountant characters",
                    ),
                    ("auditor_access", "Can view all registered characters data"),
                ),
            },
        ),
    ]



from django.utils.timezone import now
from eveuniverse.models import EveType, EveTypeMaterial

from app_utils.testing import NoSocketsTestCase

from ...models import OrePrices, Settings, get_price, get_tax, ore_calc_prices
from ..testdata.load_eveuniverse import load_eveuniverse


class TestOrePrice(NoSocketsTestCase):
    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        load_eveuniverse()

    def test_ore_calc_price(self):
        n = now()
        a = OrePrices(eve_type_id=45511, buy=10, sell=100, updated=n)  # Monazite
        a.calc_prices()
        self.assertEqual(a.buy, 10)
        self.assertEqual(a.sell, 100)
        self.assertEqual(a.raw_price, 10)
        self.assertEqual(a.refined_price, 10)
        self.assertEqual(a.taxed_price, 10)

    def test_ore_calc_prices2(self):
        n = now()
        b = EveType(eve_group_id=1920, id=16635, published=True, portion_size=1)
        b.save()
        EveTypeMaterial(
            eve_type_id=45511, material_eve_type_id=16635, quantity=3
        ).save()
        OrePrices(eve_type_id=16635, buy=20, sell=200, updated=n).save()
        a = OrePrices(eve_type_id=45511, buy=10, sell=100, updated=n)  # Monazite
        a.calc_prices()
        self.assertEqual(get_price(b), 20)
        self.assertEqual(a.buy, 10)
        self.assertEqual(a.sell, 100)
        self.assertEqual(a.raw_price, 10)
        self.assertEqual(a.refined_price, 0.54378)  # 0.9063 x 3 x 20 / 100
        self.assertEqual(a.taxed_price, 10)

    def test_ore_calc_prices3(self):
        n = now()
        b = EveType(eve_group_id=1920, id=16635, published=True, portion_size=1)
        b.save()
        EveTypeMaterial(
            eve_type_id=45511, material_eve_type_id=16635, quantity=3
        ).save()
        OrePrices(eve_type_id=16635, buy=2000, sell=200, updated=n).save()
        a = OrePrices(eve_type_id=45511, buy=10, sell=100, updated=n)  # Monazite
        a.calc_prices()
        self.assertEqual(get_price(b), 2000)
        self.assertEqual(a.buy, 10)
        self.assertEqual(a.sell, 100)
        self.assertEqual(a.raw_price, 10)
        self.assertEqual(a.refined_price, 54.378)  # 0.9063 x 3 x 2000 / 100
        self.assertEqual(a.taxed_price, 54.378)

    def test_calc_prices(self):
        n = now()
        a = OrePrices(eve_type_id=45511, buy=10, sell=100, updated=n)  # Monazite
        a.calc_prices()
        b = EveType.objects.get(id=45511)
        prices = ore_calc_prices(b, 20)
        self.assertEqual(prices[0], 200)
        self.assertEqual(prices[1], 200)
        self.assertEqual(prices[2], 200)

    def test_calc_prices2(self):
        n = now()
        b = EveType(eve_group_id=1920, id=16635, published=True, portion_size=1)
        b.save()
        EveTypeMaterial(
            eve_type_id=45511, material_eve_type_id=16635, quantity=3
        ).save()
        OrePrices(eve_type_id=16635, buy=20, sell=200, updated=n).save()
        a = OrePrices(eve_type_id=45511, buy=10, sell=100, updated=n)  # Monazite
        a.calc_prices()
        b = EveType.objects.get(id=45511)
        prices = ore_calc_prices(b, 20)
        self.assertEqual(prices[0], 200)
        self.assertEqual(prices[1], 10.8756)
        self.assertEqual(prices[2], 200)

    def test_get_tax_rates(self):
        s = Settings.load()
        s.tax_R64 = 90
        s.tax_R32 = 80
        s.tax_R16 = 70
        s.tax_R8 = 60
        s.tax_R4 = 50
        s.tax_Gasses = 40
        s.tax_Ice = 30
        s.tax_Mercoxit = 20
        s.tax_Ores = 15
        s.save()

        a = EveType.objects.get(id=45511)
        self.assertEqual(get_tax(a), 0.9)

        a = EveType.objects.get(id=45503)
        self.assertEqual(get_tax(a), 0.8)

        a = EveType.objects.get(id=45501)
        self.assertEqual(get_tax(a), 0.7)

        a = EveType.objects.get(id=45496)
        self.assertEqual(get_tax(a), 0.6)

        a = EveType.objects.get(id=45492)
        self.assertEqual(get_tax(a), 0.5)

        a = EveType.objects.get(id=28695)
        self.assertEqual(get_tax(a), 0.4)

        a = EveType.objects.get(id=16267)
        self.assertEqual(get_tax(a), 0.3)

        a = EveType.objects.get(id=11396)
        self.assertEqual(get_tax(a), 0.2)

        a = EveType.objects.get(id=1230)
        self.assertEqual(get_tax(a), 0.15)



from django.core.management.base import BaseCommand

from ...models import Character


class Command(BaseCommand):
    help = "Zeros all taxes of all characters"

    def handle(self, *args, **options):
        alltoons = Character.objects.all()
        for char in alltoons:
            balance = char.get_lifetime_taxes() - char.get_lifetime_credits()
            if balance > 0:
                char.give_credit(balance, "credit")
